@@ -208,7 +208,7 @@ int Create(char *mddev, int mdfd,
 		}
 		else {
 			ldsize = dsize;
-			dsize <<= 9;
+			ldsize <<= 9;
 		}
 		if (ldsize < MD_RESERVED_SECTORS*2LL*512LL) {
 			fprintf(stderr, Name ": %s is too small: %luK\n",
@@ -393,7 +393,7 @@ int Create(char *mddev, int mdfd,
 		if (ioctl(mdfd, RUN_ARRAY, &param)) {
 			fprintf(stderr, Name ": RUN_ARRAY failed: %s\n",
 				strerror(errno));
-			Manage_runstop(mddev, mdfd, -1);
+			Manage_runstop(mddev, mdfd, -1, 0);
 			return 1;
 		}
 		fprintf(stderr, Name ": array %s started.\n", mddev);
